import traceback
from minio import Minio


class MinioAgent:
    def __init__(self, endpoint, acs_key, secret_key, secure=True, location='default'):
        self.client = Minio(endpoint, access_key=acs_key, secret_key=secret_key, secure=secure, region=location)
        self.location = location

    def create_bucket_if_not_exists(self, bucket_name, location):
        try:
            if not self.client.bucket_exists(bucket_name):
                self.client.make_bucket(bucket_name, location=location)
            
            return True
        except Exception as e:
            traceback.print_exc()
            return False
    
    def upload_file(self, path, bucket, bucket_filename, tries=0):
        try:
            if self.create_bucket_if_not_exists(bucket, self.location):
                self.client.fput_object(bucket, bucket_filename, path)
                return True, 'minio://{}/{}'.format(bucket, bucket_filename)
            else:
                print('創建bucket失敗: {}'.format(bucket))
                return False, '創建bucket失敗'
        except Exception as e:
            if tries > 3:
                return False, e
            else:
                return self.upload_file(path, bucket, bucket_filename, tries+1)

    def upload_file_by_buffer(self, buffer, bucket, bucket_filename, tries=0):
        try:
            if self.create_bucket_if_not_exists(bucket, self.location):
                self.client.put_object(bucket, bucket_filename, buffer, len(buffer.getvalue()))
                return True, 'minio://{}/{}'.format(bucket, bucket_filename)
            else:
                return False, '創建bucket失敗'
        except Exception as e:
            if tries > 3:
                return False, e
            else:
                return self.upload_file_by_buffer(buffer, bucket, bucket_filename, tries+1)

    def download_file(self, path, bucket, bucket_filename, tries=0):
        try:
            self.client.fget_object(bucket, bucket_filename, path)
            return True, path
        except Exception as e:
            if tries > 5:
                return False, e
            else:
                return self.download_file(path, bucket, bucket_filename, tries+1)
            
    def delete(self, bucket, bucket_name, tries):
        try:
            self.client.remove_object(bucket, bucket_name)
            return True, None
        except Exception as e:
            if tries > 5:
                return False, e
            else:
                return self.delete(bucket, bucket_name, tries + 1)

    def exists(self, bucket, bucket_name):
        meta = self.client.stat_object(bucket, bucket_name)

        return meta is not None






class BackendBase:
    def __init__(self, cfg) -> None:
        self.cfg = cfg
        
    def put(self, filename, bucket, bucket_name):
        pass
    
    def put_buffer(self, buffer, bucket, bucket_name):
        pass

    async def put_async(self, filename, bucket, bucket_name):
        pass
    
    async def put_buffer_async(self, buffer, bucket, bucket_name):
        pass
    
    def get(self, filename, bucket, bucket_name):
        pass
    
    def geturl(self, entrypoint):
        return entrypoint

    async def get_async(self, filename, bucket, bucket_name):
        pass
    



import os
import json
from diskcache import Cache
from hamunafs.utils.redisutil import XRedis
from hamunafs.utils.singleton_wrapper import Singleton


class CacheManager(Singleton):
    def __init__(self, host, password, port, db=2, local_cache='../cache'):
        if not self.need_init():
            return
        self.client = XRedis(host, password, port, db=db)
        
        if local_cache is not None:
            os.makedirs(local_cache, exist_ok=True)
            self.local_cache = Cache(local_cache)
        self._inited = True
    
    def cache(self, key, val, enable_local_cache=True, expired=7200, encode=True):
        if not isinstance(val, str) and encode:
            cache_val = json.dumps(val)
        else:
            cache_val = val
        
        if enable_local_cache and hasattr(self, 'local_cache'):
            self.local_cache.set(key, cache_val, expired)
        
        self.client.set(key, cache_val, expired)

    def refresh_time(self, key, expired):
        self.client.set_expire_time(key, expired)
    
    def remove(self, key):
        self.client.delkey(key)
        self.local_cache.delete(key, retry=True)

    def get_cache(self, key, return_obj=True):
        cache_str = self.local_cache.get(key) if hasattr(self, 'local_cache') else None
        if cache_str is None:
            cache_str = self.client.get(key)
            if return_obj and cache_str is not None:
                return json.loads(cache_str)
        else:
            if isinstance(cache_str, str):
                if return_obj:
                    return json.loads(cache_str)
        return cache_str
    
    async def get_cache_async(self, key, return_obj=True):
        cache_str = self.local_cache.get(key) if hasattr(self, 'local_cache') else None
        if cache_str is None:
            cache_str = self.client.get(key)
            if return_obj and cache_str is not None:
                return json.loads(cache_str)
        else:
            if isinstance(cache_str, str):
                if return_obj:
                    return json.loads(cache_str)
        return cache_str
    
    def set_cache_reload_tag(self, tag):
        self.client.hash_set('cache_reload_tags', tag, '1')

    def lock(self, lock_key, ttl=60):
        return self.client.lock(lock_key, ttl)
    
    def zset(self, z_name, key, val):
        self.client.zadd(z_name, key, val)
    
    def zrange(self, z_name, min, max):
        return self.client.zrange(z_name, min, max)
    
    def zcount(self, z_name, min, max):
        return self.client.zcount(z_name, min, max)
    
    def zrem(self, z_name, key):
        self.client.zrem(z_name, key)
        
    def zscore(self, z_name, key):
        return self.client.zscore(z_name, key)
    
    def get_pipeline(self):
        return self.client._get_connection(pipeline=True)




"""
A clean class to work with SQLITE3 Databases in Python

Note: iud = insert, update, and delete

Examle usage (run.py file):
=================================================================================
    from DB import DB

    # Connect to DB
    myDB = DB("db-test/db.sqlite3")

    # Create a table
    dbCols = ["id integer PRIMARY KEY", "name text NOT NULL", "priority integer"]
    myDB.create_table("Users", dbCols)

    # Insert data
    myDB.iud("INSERT INTO Users (name) VALUES (?)", ("Joe",))

    # Query data
    for row in myDB.select("SELECT * FROM Users"):
        print(row)

    # Close connection
    myDB.db_close()

    ---
    Console:
    IN:  $> python run.py
    OUT: $> {'id': 1, 'name': 'Joe', 'priority': None}
=================================================================================

BY: Isaac Lehman
"""
import os
import sqlite3

class DB:
    con = None # Connection
    cur = None # Cursor


    def __init__(self, path, is_relative=True):
        """
        Setup 
        -> path to db (String)
        -> was path relative (True/False)
        """
        if is_relative:
            # get the path to the directory this script is in
            scriptdir = os.path.dirname(__file__)
            # add the relative path to the database file from there
            self.path = os.path.join(scriptdir, path)
            # make sure the path exists and if not create it
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
        else:
            self.path = path

        # Connect to the database
        self.db_connect()

    def __del__(self):
        """
        Close the connection on destruction
        """
        if self.con:
            self.db_close()
        

    def db_connect(self):
        """
        Connect to the DB
        -> Populates con and cur
        """
        self.con = sqlite3.connect(self.path, check_same_thread=False)  
        self.con.row_factory = sqlite3.Row  
        if self.con:
            self.cur = self.con.cursor()   
        else:
            raise ConnectionError(f'Unable to connect to the database at -> {self.path}')

    
    def db_close(self):
        """ 
        Close the DB connection
        """
        if self.con:
            self.con.close()


    def create_table(self, table_name, table_cols):
        """
        Create a table in the DB
        -> Name of table (String)
        -> Array of column statements (Array[String])
            ex. ["id integer PRIMARY KEY", "name text NOT NULL", ...]
        """
        create_statement = f'CREATE TABLE IF NOT EXISTS {table_name} ('

        num_rows = len(table_cols)
        index = 0
        for col in table_cols:
            index += 1
            if index == num_rows:
                create_statement += f' {col}'
            else:
                create_statement += f' {col},'

        create_statement += ')'

        self.cur.execute(create_statement)

    
    def add_column(self, table, new_col):
        """
        Add a column to a table in the DB
        -> Name of table (String)
        -> Name of the new column (String)
            ex. last_name text NOT NULL
        """
        create_statement = f'ALTER TABLE {table} ADD COLUMN {new_col}'
        self.cur.execute(create_statement)
        


    def select(self, statement, args=None):
        """
        Select data from the DB
        -> The SQL statment (String)
        -> A tuple of arguments to fill a prepared statement (tuple)
            Place ? where they go
            ex. "SELECT * FROM Users WHERE id=? and name=?", (id, name)  
        """
        results = None
        if args:
            results = self.cur.execute(statement, args)
        else:
            results = self.cur.execute(statement)

        if results:
            return [dict(row) for row in results]
        else:
            return None


    def iud(self, statement, args=None):
        """
        Insert/Update/Delet data into the DB
        -> The SQL statment (String)
        -> A tuple of arguments to fill a prepared statement (tuple)
            Place ? where they go
            ex. "INSERT INTO Users (id, name) VALUES (?, ?)", (id, name)  
        """

        if args:
            self.cur.execute(statement, args)
        else:
            self.cur.execute(statement)
        
        self.con.commit()


    def sql(self, statement, args=None, commit=False):
        """
        Execute arbitrary SQL
        -> The SQL statment (String)
        -> A tuple of arguments to fill a prepared statement (tuple)
            Place ? where they go
        -> Whether or not to commit to the DB (True/False)
        """
        if args:
            results = self.cur.execute(statement, args)
        else:
            results = self.cur.execute(statement)

        if commit:
            self.con.commit()
        
        return results




import traceback
from gnsq import Producer, Consumer, Message, NsqdHTTPClient
from ansq import open_connection
import asyncio
import orjson
import time

from hamunafs.utils.singleton_wrapper import Singleton


class MQManager(Singleton):
    def __init__(self, host, port, async_mq=False, init=True):
        if not self.need_init():
            return
        self.host = host
        self.port = port
        self.async_mq = async_mq
        self.inited = False
        if init:
            self._init()
        self._inited = True

    def _init(self):
        if self.async_mq:
            self.mq = asyncio.get_event_loop().run_until_complete(open_connection(self.host, self.port))
        else:
            self.addr = '{}:{}'.format(self.host, self.port)
            self.producer = Producer(nsqd_tcp_addresses=[self.addr])
            self.producer.start()
        self.inited = True

    async def get_mq_conn(self):
        return await open_connection(self.host, self.port)
            
    def _encode_message(self, message):
        if isinstance(message, str):
            message = message.encode('utf-8')
        else:
            message = orjson.dumps(message)
        return message
            
    def publish(self, topic, message, multi=False):
        try:
            if multi and isinstance(message, list):
                message = list(map(self._encode_message, message))
            else:
                message = self._encode_message(message)
            
            if multi:
                pub_func = self.producer.multipublish
            else:
                pub_func = self.producer.publish
            
            ret = pub_func(topic, message).decode()
        except Exception as e:
            traceback.print_exc()
            ret = False
        max_t = 5
        t = 0
        while ret != 'OK' and t < max_t: 
            try:
                ret = pub_func(topic, message).decode()
            except:
                ret = 'ERR'
            if ret == 'ERR':
                time.sleep(1)
            t += 1
        ret = ret == 'OK'
        if ret:
            print('nsq publish success!')
        return ret

    async def async_publish(self, topic, message, mq=None):
        try:
            if isinstance(message, str):
                message = message.encode('utf-8')
            elif isinstance(message, bytes):
                pass
            else:
                message = orjson.dumps(message)
            print('publishing')

            if mq is None:
                if self.inited:
                    mq = self.mq
                else:
                    self._init()
                    mq = self.mq

            ret = await mq.pub(topic, message)
            max_t = 5
            t = 0
            while not ret and t < max_t: 
                try:
                    ret = await mq.pub(topic, message)
                except:
                    ret = False
                t += 1
            return ret
        except Exception as e:
            print(traceback.print_exc())
            ret = False

if __name__ == "__main__":
    manager = MQManager('kafka.ai.hamuna.club', 34150, async_mq=True)
    
    loop = asyncio.get_event_loop()
    loop.run_until_complete(manager._init_async())
    
    for i in range(100):
        ret = loop.run_until_complete(manager.async_publish('test', 'test'))
        # ret = manager.publish('test', 'test'.encode('utf-8'))
        print(ret)